import * as assert from 'assert';
import json0 from 'ot-json0';
import json1 from 'ot-json1';
import textUnicode from 'ot-text-unicode';
import otDiff from 'json0-ot-diff';
import jsondiff from 'json0-ot-diff';
import diffMatchPatch from 'diff-match-patch';
import { EditorState, ChangeSet } from '@codemirror/state';
import { EditorView } from '@codemirror/view';
import {
  changesToOpJSON0,
  changesToOpJSON1,
  opToChangesJSON0,
  opToChangesJSON1,
} from '../src/index';

const atPath = (obj, path) => path.reduce((d, key) => d[key], obj);
const clone = (obj) => JSON.parse(JSON.stringify(obj));
const diffJSON0 = (a, b) => jsondiff(a, b, diffMatchPatch);
const diffJSON1 = (a, b) => jsondiff(a, b, diffMatchPatch, json1, textUnicode);

// Verifies that ops and transaction match in terms of behavior,
// and that the translation between ops and transaction holds in both directions.
export const verify = (options) => {
  const { before, after, changes, opJSON0, opJSON1, path = [] } = options;

  it('JSON0 op should match computed diff', () => {
    if (!opJSON0) {
      console.log(`opJSON0: ${JSON.stringify(diffJSON0(before, after))},`);
      process.exit();
    }
    assert.deepEqual(opJSON0, diffJSON0(before, after));
  });

  it('JSON1 op should match computed diff', () => {
    if (!opJSON1) {
      console.log(`opJSON1: ${JSON.stringify(diffJSON1(before, after))},`);
      process.exit();
    }
    assert.deepEqual(opJSON1, diffJSON1(before, after));
  });

  it('JSON0 applied op should match expected text', () => {
    assert.deepEqual(after, json0.type.apply(clone(before), opJSON0));
  });

  it('JSON0 inverted applied op should match expected text', () => {
    assert.deepEqual(
      before,
      json0.type.apply(clone(after), json0.type.invert(opJSON0))
    );
  });

  it('JSON1 applied op should match expected text', () => {
    assert.deepEqual(after, json1.type.apply(clone(before), opJSON1));
  });

  it('JSON1 inverted applied op should match expected text', () => {
    const opJSON1Invertible = json1.type.makeInvertible(opJSON1, before);
    assert.deepEqual(
      before,
      json1.type.apply(clone(after), json1.type.invert(opJSON1Invertible))
    );
  });

  it('opToChangesJSON0', () => {
    if (!changes) {
      console.log(
        `changes: ${JSON.stringify(opToChangesJSON0(opJSON0))}, // from json0`
      );
      console.log(
        `changes: ${JSON.stringify(opToChangesJSON1(opJSON1))}, // from json1`
      );
      process.exit();
    }
    assert.deepEqual(opToChangesJSON0(opJSON0), changes);
  });

  it('opToChangesJSON1', () => {
    assert.deepEqual(opToChangesJSON1(opJSON1), changes);
  });

  it('changesToOpJSON0', () => {
    const state = EditorState.create({ doc: atPath(before, path) });
    const changeSet = ChangeSet.of(changes, before.length);
    //console.log();
    //console.log(changes);
    //console.log(JSON.stringify(changeSet.toJSON()));
    assert.deepEqual(changesToOpJSON0(path, changeSet, state.doc), opJSON0);
  });

  it('changesToOpJSON1', () => {
    const state = EditorState.create({ doc: atPath(before, path) });
    const changeSet = ChangeSet.of(changes, before.length);
    assert.deepEqual(changesToOpJSON1(path, changeSet, state.doc), opJSON1);
  });

  it('applied changes should match expected text', () => {
    const state = EditorState.create({ doc: atPath(before, path) });
    const expected = atPath(after, path);
    const actual = state.update({ changes }).state.doc.sliceString(0);
    assert.equal(expected, actual);
  });
};
